module multipass
  use const_and_precisions, only : wp_, zero, half, one, degree, czero
  use beamdata, only : dst, nray
  use gray_params, only : ipass
  use polarization, only : pol_limit, field_to_ellipse
  use reflections, only : wall_refl
  use equilibrium, only : bfield
  
  implicit none

  integer, save :: nbeam_max                                                               ! max n of beams active at a time
  integer, save :: nbeam_tot                                                               ! total n of beams
  
contains

  subroutine plasma_in(i, x, N, Bres, sox, cpl, psi, chi, iop, ext, eyt, perfect)
    ! Computes the ray polarisation and power couplings when it enteres the plasma
    use const_and_precisions, only: cm

    ! subroutine arguments
    integer,      intent(in)             :: i              ! ray index
    real(wp_),    intent(in)             :: x(3), N(3)     ! position, refactive index
    real(wp_),    intent(in)             :: Bres           ! resonant B field
    integer,      intent(in)             :: sox            ! sign of polarisation mode: -1 ⇒ O, +1 ⇒ X
    real(wp_),    intent(out)            :: cpl(2)         ! power coupling vector (O, X)
    real(wp_),    intent(out)            :: psi, chi       ! polarisation ellipse angles
    integer,      intent(inout), pointer :: iop(:)         ! inside/outside plasma flag
    complex(wp_), intent(inout), pointer :: ext(:), eyt(:) ! ray polarisation vector (e_x, e_y)
    logical,      intent(in)             :: perfect        ! whether to assume perfect coupling

    ! local variables
    real(wp_) :: R, z, cosphi, sinphi, B_phi, B_R, B_z
    real(wp_) :: B(3)
    real(wp_) :: c
    complex(wp_) :: e_mode(2), e_ray(2)

    ! Update the inside/outside flag
    iop(i) = iop(i) + 1

    ! Compute B in cartesian coordinates
    R = norm2(x(1:2)) * cm
    z = x(3) * cm
    cosphi = x(1)/R * cm
    sinphi = x(2)/R * cm
    call bfield(R, z, B_phi, B_R, B_z)
    B(1) = B_R*cosphi - B_phi*sinphi
    B(2) = B_R*sinphi + B_phi*cosphi
    B(3) = B_z

    ! Get the polarisation vector of the given mode
    call pol_limit(N, B, Bres, sox, e_mode(1), e_mode(2))

    if(i == 1) then
      ! For the central ray, compute the polarization ellipse
      call field_to_ellipse(e_mode(1), e_mode(2), psi, chi)
    else
      psi = 0
      chi = 0
    end if

    if (perfect) then
      ! Ignore the given vector and use the expected one
      ! Note: this will give 100% coupling to the current mode
      ext(i) = e_mode(1)
      eyt(i) = e_mode(2)
    end if

    ! Compute the power coupling with the current mode
    e_ray = [ext(i), eyt(i)]
    c = abs(dot_product(e_mode, e_ray))**2

    ! Store both O and X couplings, in this order
    c = merge(c, 1-c, sox == -1)
    cpl = [c, 1-c]
  end subroutine plasma_in


  subroutine plasma_out(i,xv,anv,bres,sox,iop,ext,eyt)                                     ! ray exits plasma
    implicit none
! arguments
    integer, intent(in) :: i                                                               ! ray index
    real(wp_), dimension(3), intent(in) :: xv,anv
    real(wp_), intent(in) :: bres
    integer,   intent(in) :: sox
    integer, dimension(:), intent(inout), pointer :: iop                                   ! in/out plasma flag
    complex(wp_), dimension(:), intent(out), pointer :: ext,eyt
! local variables
    real(wp_) :: rm,csphi,snphi,bphi,br,bz
    real(wp_), dimension(3) :: bv,xmv
!     
    iop(i)=iop(i)+1                                                                        ! in->out
    
    xmv=xv*0.01_wp_ ! convert from cm to m
    rm=sqrt(xmv(1)**2+xmv(2)**2)
    csphi=xmv(1)/rm
    snphi=xmv(2)/rm
    call bfield(rm,xmv(3),bphi,br,bz)
    bv(1)=br*csphi-bphi*snphi
    bv(2)=br*snphi+bphi*csphi
    bv(3)=bz
    call pol_limit(anv,bv,bres,sox,ext(i),eyt(i))                                          ! polarization at plasma exit
    
  end subroutine plasma_out
! ------------------------------
!   subroutine wall_in(i)                                                                    ! ray enters vessel
!     implicit none
!     integer, intent(in) :: i                                                               ! ray index
!   
!     iow(i)=iow(i)+1
!   end subroutine wall_in
! ------------------------------
  subroutine wall_out(i,ins,xv,anv,bres,sox,psipol1,chipol1,iow,iop,ext,eyt)               ! ray exits vessel
    implicit none
! arguments
    integer, intent(in) :: i                                                               ! ray index
    logical, intent(in) :: ins                                                             ! inside plasma? (ins=1 plasma/wall overlap)
    real(wp_), dimension(3), intent(inout) :: xv,anv
    real(wp_), intent(in) :: bres
    integer,   intent(in) :: sox
    real(wp_), intent(out) :: psipol1,chipol1
    integer, dimension(:), intent(inout), pointer :: iow,iop                               ! in/out vessel and plasma flags
    complex(wp_), dimension(:), intent(inout), pointer :: ext,eyt
! local variables
    integer :: irfl
    real(wp_), dimension(3) :: xvrfl,anvrfl,walln
    complex(wp_) :: ext1,eyt1
!   
    iow(i)=iow(i)+1                                                                        ! out->in
    if(ins) call plasma_out(i,xv,anv,bres,sox,iop,ext,eyt)                                 ! plasma-wall overlapping
    call wall_refl(xv-dst*anv,anv,ext(i),eyt(i),xvrfl,anvrfl,ext1,eyt1,walln,irfl)         ! ray reflects at wall
    ext(i) = ext1                                                                          ! save parameters at wall reflection
    eyt(i) = eyt1
    xv  = xvrfl
    anv = anvrfl

    if(i == 1) then                                                                        ! polarization angles at wall reflection for central ray
      call field_to_ellipse(ext1, eyt1, psipol1, chipol1)
    else
      psipol1 = zero
      chipol1 = zero
    end if

  end subroutine wall_out


  subroutine initbeam(params, i, iroff, iboff, iwait, stv, dst, jphi_beam, &
                      pins_beam, currins_beam, dpdv_beam, jcd_beam)
    ! Initialises the beam variables at the start of the beam propagation

    use gray_params, only : gray_parameters
    use logger,      only : log_warning

    implicit none

    ! subroutine arguments
    type(gray_parameters), intent(in)  :: params
    integer,               intent(in)  :: i             ! beam index
    logical, pointer,      intent(in)  :: iroff(:,:)    ! global ray status (F = active, T = inactive)
    logical,               intent(out) :: iboff         ! beam status (F = active, T = inactive)
    logical, pointer,      intent(out) :: iwait(:)
    real(wp_), pointer,    intent(out), dimension(:) :: &
       jphi_beam, pins_beam,  currins_beam, dpdv_beam, jcd_beam, stv, dst

    ! local variables
    character(256) :: msg  ! buffer for formatting log messages

    iboff = .false.
    iwait = iroff(:,i)  ! copy current beam status from the global one

    if (all(iwait)) then
      ! no rays active => stop beam
      iboff = .true.
    else if (any(iwait)) then
      ! only some rays active
      write (msg,'("  beam ",g0,": only some rays are active!")') i
      call log_warning(msg, mod='multipass', proc='initbeam')
    end if

    stv = zero                   ! starting ray parameter (s, c⋅t, S_R)
    dst = params%raytracing%dst  ! starting step size (ds, c⋅dt, dS_R)

    ! 1D beam profiles
    jphi_beam    = zero
    pins_beam    = zero
    currins_beam = zero
    dpdv_beam    = zero
    jcd_beam     = zero
  end subroutine initbeam


  subroutine initmultipass(i,iox,iroff,yynext,yypnext,yw0,ypw0,stnext,p0ray, &
                           taus,tau1,etau1,cpls,cpl1,lgcpl1,psipv,chipv)                   ! initialization before pass loop
    implicit none
! arguments
    integer, intent(in) :: i, iox                                                          ! ipol, mode active on 1st pass
    logical, dimension(:,:), intent(out), pointer :: iroff                                 ! global ray status (F = active, T = inactive)
    real(wp_), dimension(:), intent(out), pointer :: p0ray,tau1,etau1,cpl1,lgcpl1, &
       psipv,chipv
    real(wp_), dimension(:,:), intent(out), pointer :: yw0,ypw0,stnext,taus,cpls
    real(wp_), dimension(:,:,:), intent(out), pointer :: yynext,yypnext
!   
    iroff = .false.                                                                        ! global ray status (F = active, T = inactive)
    if(i.eq.0) call turnoffray(0,1,3-iox,iroff)                                            ! ipol = 0 => stop other mode (iox=1/2 -> stop ib=2/1 at first pass)
    yynext  = zero                                                                         ! starting beam coordinates (1)
    yypnext = zero                                                                         ! starting beam coordinates (2)
    yw0     = zero                                                                         ! temporary beam coordinates (1)
    ypw0    = zero                                                                         ! temporary beam coordinates (2)
    stnext  = zero                                                                         ! starting beam step
    p0ray   = zero                                                                         ! starting beam power
    taus    = zero                                                                         ! beam tau from previous passes
    tau1    = zero
    etau1   = one
    cpls    = one                                                                          ! beam coupling from previous passes
    cpl1    = one
    lgcpl1  = zero
    psipv   = zero                                                                         ! psi polarization angle at vacuum-plasma boundary
    chipv   = zero                                                                         ! chi polarization angle at vacuum-plasma boundary
  end subroutine initmultipass
! ------------------------------
  subroutine turnoffray(jk,ip,ib,iroff)                                                    ! turn off ray propagation
    implicit none
! arguments
    integer, intent(in) :: jk, ip, ib                                                      ! ray (0=all rays), pass, beam indexes
    logical, dimension(:,:), intent(out), pointer :: iroff                                 ! global ray status (F = active, T = inactive)
! local variables
    integer :: ipx, i1, i2
!   
    if(jk==0) then                                                                         ! stop all rays
      do ipx=ip,ipass                                                                      ! from pass ip to last pass
        i1 = 2**ipx-2+2**(ipx-ip)*(ib-1)+1                                                 ! first derived beam at pass ipx
        i2 = 2**ipx-2+2**(ipx-ip)*ib                                                       ! last derived beam at pass ipx (i1=i2 for ipx=ip)
        iroff(:,i1:i2) = .true.
      end do
    else                                                                                   ! only stop ray jk
      do ipx=ip,ipass
        i1 = 2**ipx-2+2**(ipx-ip)*(ib-1)+1
        i2 = 2**ipx-2+2**(ipx-ip)*ib
        iroff(jk,i1:i2) = .true.
      end do
    end if
    
  end subroutine turnoffray


  subroutine alloc_multipass(&
    dim, iwait, iroff, iop, iow, yynext, yypnext, yw0, ypw0, stnext,   &
    stv, dst, p0ray, taus, tau1, etau1, cpls, cpl1, lgcpl1, jphi_beam, &
    pins_beam, currins_beam, dpdv_beam, jcd_beam, psipv, chipv)

    implicit none

    ! subroutine arguments
    integer,                            intent(in)  :: dim
    logical,   pointer, dimension(:),   intent(out) :: iwait
    logical,   pointer, dimension(:,:), intent(out) :: iroff
    integer,   pointer, dimension(:),   intent(out) :: iop, iow
    real(wp_), pointer, dimension(:),   intent(out) :: &
       jphi_beam, pins_beam, currins_beam, dpdv_beam,  &
       jcd_beam, stv, dst, tau1, etau1, cpl1, lgcpl1,  &
       p0ray, psipv, chipv
    real(wp_), pointer, dimension(:, :),    intent(out) :: taus, cpls, stnext
    real(wp_), pointer, dimension(:, :),    intent(out) :: yw0, ypw0
    real(wp_), pointer, dimension(:, :, :), intent(out) :: yynext, yypnext
    
    call dealloc_multipass(&
      iwait, iroff, iop, iow, yynext, yypnext, yw0, ypw0, stnext,        &
      stv, dst, p0ray, taus, tau1, etau1, cpls, cpl1, lgcpl1, jphi_beam, &
      pins_beam, currins_beam, dpdv_beam, jcd_beam, psipv, chipv)

    
    nbeam_max = 2**ipass        ! max n of beams active at a time
    nbeam_tot = 2**(ipass+1)-2  ! total n of beams
    
    allocate(iwait(nray))
    allocate(iroff(nray, nbeam_tot))
    allocate(iop(nray))
    allocate(iow(nray))

    allocate(yynext(6, nray, nbeam_max-2))
    allocate(yypnext(6, nray, nbeam_max-2))

    allocate(yw0(6, nray))
    allocate(ypw0(6, nray))

    allocate(stnext(nray, nbeam_tot))
    allocate(stv(nray))
    allocate(dst(nray))

    allocate(p0ray(nray))
    allocate(taus(nray, nbeam_tot))
    allocate(tau1(nray))
    allocate(etau1(nray))

    allocate(cpls(nray, nbeam_tot))
    allocate(cpl1(nray))
    allocate(lgcpl1(nray))
    allocate(jphi_beam(dim))

    allocate(pins_beam(dim))
    allocate(currins_beam(dim))
    allocate(dpdv_beam(dim))
    allocate(jcd_beam(dim))

    allocate(psipv(0:nbeam_tot))
    allocate(chipv(0:nbeam_tot))

  end subroutine alloc_multipass


  subroutine dealloc_multipass(&
    iwait, iroff, iop, iow, yynext, yypnext, yw0, ypw0, stnext,        &
    stv, dst, p0ray, taus, tau1, etau1, cpls, cpl1, lgcpl1, jphi_beam, &
    pins_beam, currins_beam, dpdv_beam, jcd_beam, psipv, chipv)

    implicit none

    logical,   pointer, dimension(:),   intent(out) :: iwait
    logical,   pointer, dimension(:,:), intent(out) :: iroff
    integer,   pointer, dimension(:),   intent(out) :: iop, iow
    real(wp_), pointer, dimension(:),   intent(out) :: &
       jphi_beam, pins_beam, currins_beam, dpdv_beam,  &
       jcd_beam, stv, dst, tau1, etau1, cpl1, lgcpl1,  &
       p0ray, psipv, chipv
    real(wp_), pointer, dimension(:, :),    intent(out) :: taus, cpls, stnext
    real(wp_), pointer, dimension(:, :),    intent(out) :: yw0, ypw0
    real(wp_), pointer, dimension(:, :, :), intent(out) :: yynext, yypnext

    if (associated(iwait))         deallocate(iwait)
    if (associated(iroff))         deallocate(iroff)
    if (associated(iop))           deallocate(iop)
    if (associated(iow))           deallocate(iow)

    if (associated(yynext))        deallocate(yynext)
    if (associated(yypnext))       deallocate(yypnext)

    if (associated(yw0))           deallocate(yw0)
    if (associated(ypw0))          deallocate(ypw0)

    if (associated(stnext))        deallocate(stnext)
    if (associated(stv))           deallocate(stv)
    if (associated(dst))           deallocate(dst)

    if (associated(p0ray))         deallocate(p0ray)
    if (associated(taus))          deallocate(taus)
    if (associated(tau1))          deallocate(tau1)
    if (associated(etau1))         deallocate(etau1)

    if (associated(cpls))          deallocate(cpls)
    if (associated(cpl1))          deallocate(cpl1)
    if (associated(lgcpl1))        deallocate(lgcpl1)
    if (associated(jphi_beam))     deallocate(jphi_beam)

    if (associated(pins_beam))     deallocate(pins_beam)
    if (associated(currins_beam))  deallocate(currins_beam)
    if (associated(dpdv_beam))     deallocate(dpdv_beam)
    if (associated(jcd_beam))      deallocate(jcd_beam)

    if (associated(psipv))         deallocate(psipv)
    if (associated(chipv))         deallocate(chipv)
  end subroutine dealloc_multipass

end module multipass
